##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

"""
Jinja2 template filter

To use, add the following to your configuration file:

.. code-block:: yaml

   filter: exhibition.filters.jinja2
"""

from datetime import datetime, timezone

from jinja2 import Environment, FileSystemLoader, contextfilter
from jinja2.exceptions import TemplateRuntimeError
from jinja2.ext import Extension
from jinja2.nodes import CallBlock, Const, ContextReference
from markdown import markdown as md_func
from pypandoc import convert_text as pandoc_func
from typogrify.templatetags import jinja_filters as typogrify_filters

EXTENDS_TEMPLATE_TEMPLATE = """{%% extends "%s" %%}
"""
START_BLOCK_TEMPLATE = """{%% block %s %%}
"""
END_BLOCK_TEMPLATE = """{% endblock %}
"""

DEFAULT_GLOB = "*.html"

NODE_TMPL_VAR = "node"

DEFAULT_MD_KWARGS = {
    "output_format": "html5",
}

DEFAULT_PANDOC_KWARGS = {
    "to": "html",
}


def metasort(nodes, key=None, reverse=False):
    """
    Sorts a list of nodes based on keys found in their meta objects
    """
    def key_func(node):
        return node.meta[key]

    return sorted(nodes, key=key_func, reverse=reverse)


def metaselect(nodes, key):
    for n in nodes:
        if n.meta.get(key):
            yield n


def metareject(nodes, key):
    for n in nodes:
        if not n.meta.get(key):
            yield n


@contextfilter
def markdown(ctx, text):
    kwargs = DEFAULT_MD_KWARGS.copy()
    node = ctx[NODE_TMPL_VAR]

    kwargs.update(node.meta.get("markdown_config", {}))

    return md_func(text, **kwargs)


@contextfilter
def pandoc(ctx, text, fmt=None):
    kwargs = DEFAULT_PANDOC_KWARGS.copy()
    node = ctx[NODE_TMPL_VAR]

    kwargs.update(node.meta.get("pandoc_config", {}))

    if fmt is not None:
        kwargs["format"] = fmt

    return pandoc_func(text, **kwargs)


class RaiseError(Extension):
    """
    Raise an exception during template rendering:

    .. code-block:: jinja

       {% raise "This is an error" %}
    """
    tags = set(["raise"])

    def _raise_error(self, message, caller):
        raise TemplateRuntimeError(message)

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        message = parser.parse_expression()

        return CallBlock(
            self.call_method('_raise_error', [message], lineno=lineno),
            [], [], []).set_lineno(lineno)


class Mark(Extension):
    """
    Marks a section for use later:

    .. code-block:: html+jinja

       {% mark intro %}
       <p>My Intro</p>
       {% endmark %}

       <p>Some more text</p>

    This can then be referenced via :attr:`Node.marks`.
    """
    tags = set(['mark'])

    def parse(self, parser):
        token = next(parser.stream)
        lineno = token.lineno
        tag = token.value
        name = next(parser.stream).value
        body = parser.parse_statements(["name:end%s" % tag], drop_needle=True)
        return CallBlock(self.call_method('_render_output', args=[ContextReference(), Const(name)]),
                         [], [], body).set_lineno(lineno)

    def _render_output(self, ctx, name, caller):
        """
        Assign the marked content to :attr:`Node.marks`
        """
        out = caller()
        ctx[NODE_TMPL_VAR].marks[name] = out
        return out


def content_filter(node, content):
    """
    This is the actual content filter called by :class:`exhibition.main.Node`
    on appropiate nodes.

    :param node:
        The node being rendered
    :param content:
        The content of the node, stripped of any YAML frontmatter
    """
    env = Environment(
        loader=FileSystemLoader(node.meta["templates"]),
        extensions=[RaiseError, Mark],
        autoescape=True,
    )
    env.filters["pandoc"] = pandoc
    env.filters["markdown"] = markdown
    env.filters["metasort"] = metasort
    env.filters["metaselect"] = metaselect
    env.filters["metareject"] = metareject
    typogrify_filters.register(env)

    parts = []

    if node.meta.get("extends"):
        parts.append(EXTENDS_TEMPLATE_TEMPLATE % node.meta["extends"])

    if node.meta.get("default_block"):
        parts.extend([
            START_BLOCK_TEMPLATE % node.meta["default_block"],
            content,
            END_BLOCK_TEMPLATE,
        ])
    else:
        parts.append(content)

    content = "".join(parts)

    template = env.from_string(content)

    return template.render({
        NODE_TMPL_VAR: node,
        "time_now": datetime.now(timezone.utc),
    })



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from io import StringIO
from tempfile import NamedTemporaryFile
from unittest import TestCase, mock

from ruamel.yaml import YAML

from exhibition.config import SITE_YAML_PATH, Config

YAML_DATA = """
sitename: bob
thingy:
    - one
    - two
    - three
"""


class ConfigTestCase(TestCase):
    def test_load_str(self):
        settings = Config(YAML_DATA)

        self.assertEqual(len(settings), 2)
        self.assertEqual(settings["sitename"], "bob")
        self.assertEqual(settings["thingy"], ["one", "two", "three"])

    def test_load_file(self):
        yaml_file = StringIO(YAML_DATA)
        settings = Config(yaml_file)

        self.assertEqual(len(settings), 2)
        self.assertEqual(settings["sitename"], "bob")
        self.assertEqual(settings["thingy"], ["one", "two", "three"])

    def test_load_dict(self):
        yaml_dict = dict(YAML(typ="safe").load(YAML_DATA))
        settings = Config(yaml_dict)

        self.assertEqual(len(settings), 2)
        self.assertEqual(settings["sitename"], "bob")
        self.assertEqual(settings["thingy"], ["one", "two", "three"])

    def test_load_AssertionError(self):
        # raise an AssertionError if it's not a dict or file-like object
        with self.assertRaises(AssertionError):
            Config(12)

    def test_no_load(self):
        settings = Config()
        self.assertEqual(len(settings), 0)

    def test_from_path(self):
        with NamedTemporaryFile() as yaml_file:
            yaml_file.write(YAML_DATA.encode())
            yaml_file.file.flush()  # just to be sure

            settings = Config.from_path(yaml_file.name)

            self.assertEqual(len(settings), 2)
            self.assertEqual(settings["sitename"], "bob")
            self.assertEqual(settings["thingy"], ["one", "two", "three"])

    def test_getitem(self):
        settings = Config(YAML_DATA)

        self.assertEqual(settings["sitename"], "bob")
        self.assertEqual(settings["thingy"], ["one", "two", "three"])

        self.assertEqual(settings.get("sitename"), "bob")
        self.assertEqual(settings.get("sitetitle"), None)

        with self.assertRaises(KeyError):
            settings["sitetitle"]

    def test_getitem_with_parent(self):
        parent = Config({"test": True})
        settings = Config(YAML_DATA, parent=parent, node=mock.Mock())

        self.assertEqual(settings["sitename"], "bob")
        self.assertEqual(settings["thingy"], ["one", "two", "three"])
        self.assertEqual(settings["test"], True)

        self.assertEqual(settings.get("sitename"), "bob")
        self.assertEqual(settings.get("test"), True)
        self.assertEqual(settings.get("sitetitle"), None)
        self.assertEqual(parent.get("sitename"), None)

        with self.assertRaises(KeyError):
            settings["sitetitle"]

    def test_setitem(self):
        settings = Config(YAML_DATA)

        settings["sitename"] = "mysite"

        self.assertEqual(settings["sitename"], "mysite")
        self.assertEqual(settings["thingy"], ["one", "two", "three"])

    def test_setitem_with_parent(self):
        parent = Config({"test": True})
        settings = Config(YAML_DATA, parent=parent, node=mock.Mock())

        settings["sitename"] = "mysite"
        settings["test"] = False

        self.assertEqual(settings["sitename"], "mysite")
        self.assertEqual(settings["thingy"], ["one", "two", "three"])
        self.assertEqual(settings["test"], False)
        self.assertEqual(parent["test"], True)

        with self.assertRaises(KeyError):
            parent["sitename"]

    def test_keys(self):
        settings = Config(YAML_DATA)

        self.assertCountEqual(list(settings.keys()), ["sitename", "thingy"])

    def test_keys_with_parent(self):
        parent = Config({"test": True, "sitename": "me"})
        settings = Config(YAML_DATA, parent=parent, node=mock.Mock())

        self.assertCountEqual(list(settings.keys()), ["sitename", "thingy", "test"])
        self.assertCountEqual(list(parent.keys()), ["test", "sitename"])

    def test_iter(self):
        settings = Config(YAML_DATA)

        self.assertCountEqual(list(settings), ["sitename", "thingy"])

    def test_iter_with_parent(self):
        parent = Config({"test": True, "sitename": "me"})
        settings = Config(YAML_DATA, parent=parent, node=mock.Mock())

        self.assertCountEqual(list(settings.keys()), ["sitename", "thingy", "test"])
        self.assertCountEqual(list(parent.keys()), ["test", "sitename"])

    def test_values(self):
        settings = Config(YAML_DATA)

        self.assertCountEqual(list(settings.values()), ["bob", ["one", "two", "three"]])

    def test_values_with_parent(self):
        parent = Config({"test": True})
        settings = Config(YAML_DATA, parent=parent, node=mock.Mock())

        self.assertCountEqual(list(settings.values()), ["bob", ["one", "two", "three"], True])
        self.assertCountEqual(list(parent.values()), [True])

    def test_items(self):
        settings = Config(YAML_DATA)

        self.assertCountEqual(list(settings.items()),
                              [("sitename", "bob"), ("thingy", ["one", "two", "three"])])

    def test_items_with_parent(self):
        parent = Config({"test": True})
        settings = Config(YAML_DATA, parent=parent, node=mock.Mock())

        self.assertCountEqual(list(settings.items()),
                              [("sitename", "bob"), ("thingy", ["one", "two", "three"]),
                               ("test", True)])
        self.assertCountEqual(list(parent.items()), [("test", True)])

    def test_contains(self):
        settings = Config(YAML_DATA)

        self.assertIn("sitename", settings)
        self.assertIn("thingy", settings)
        self.assertNotIn("test", settings)

    def test_contains_with_parent(self):
        parent = Config({"test": True})
        settings = Config(YAML_DATA, parent=parent, node=mock.Mock())

        self.assertIn("test", settings)
        self.assertIn("sitename", settings)
        self.assertIn("thingy", settings)

        self.assertIn("test", parent)
        self.assertNotIn("sitename", parent)
        self.assertNotIn("thingy", parent)

    def test_len(self):
        settings = Config(YAML_DATA)

        self.assertEqual(len(settings), 2)

    def test_len_with_parent(self):
        parent = Config({"test": True})
        settings = Config(YAML_DATA, parent=parent, node=mock.Mock())

        self.assertEqual(len(settings), 3)
        self.assertEqual(len(parent), 1)

    def test_parent_and_node(self):
        with self.assertRaises(AssertionError):
            Config({}, parent=Config({}))

        with self.assertRaises(AssertionError):
            Config({}, node=mock.Mock())

        # these should be fine
        Config({})
        Config({}, parent=Config({}), node=mock.Mock())

    def test_get_name_for_root_config(self):
        config = Config({})
        self.assertEqual(config.get_name(), SITE_YAML_PATH)

    def test_get_name_for_node_config(self):
        node = mock.Mock()
        node.full_path = "this-file.html"
        config = Config({}, parent=mock.Mock(), node=node)
        self.assertEqual(config.get_name(), node.full_path)

    def test_KeyError_contains_name_of_child(self):
        node = mock.Mock()
        node.full_path = "this-file.html"
        config = Config({}, parent=Config({}), node=node)

        with self.assertRaises(KeyError) as exp:
            config["some key"]

        self.assertIn("this-file.html", str(exp.exception))

    def test_KeyError_contains_name_of_site_yaml(self):
        config = Config({})

        with self.assertRaises(KeyError) as exp:
            config["some key"]

        self.assertIn(SITE_YAML_PATH, str(exp.exception))

    def test_copy(self):
        parent = Config({})
        settings = Config(YAML_DATA, parent=parent, node=mock.Mock())
        copied = settings.copy()

        self.assertNotEqual(settings, copied)
        self.assertEqual([i for i in settings.keys()], [i for i in copied.keys()])
        self.assertEqual(settings.parent, copied.parent)

        settings["bob"] = "hello"
        self.assertNotEqual([i for i in settings.keys()], [i for i in copied.keys()])

    def test_repr(self):
        settings = Config({})
        # just a smoke test to make sure it doesn't blow up
        self.assertTrue(isinstance(repr(settings), str))
        self.assertTrue(repr(settings))



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from filecmp import dircmp
from tempfile import TemporaryDirectory
from unittest import TestCase, mock
import logging
import pathlib

from click.testing import CliRunner

from exhibition import command, config
import exhibition


class CommandTestCase(TestCase):
    @mock.patch("exhibition.command.utils.gen")
    @mock.patch("exhibition.command.config.Config.from_path", return_value=config.Config())
    def test_gen(self, config_mock, gen_mock):
        runner = CliRunner()
        result = runner.invoke(command.exhibition, ["gen"])

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(gen_mock.call_count, 1)
        self.assertEqual(gen_mock.call_args, ((config_mock.return_value,), {}))

        self.assertEqual(config_mock.call_args, ((config.SITE_YAML_PATH,), {}))

    @mock.patch("exhibition.command.utils.serve", return_value=(mock.Mock(), mock.Mock()))
    @mock.patch("exhibition.command.config.Config.from_path", return_value=config.Config())
    def test_serve(self, config_mock, serve_mock):
        runner = CliRunner()
        result = runner.invoke(command.exhibition, ["serve"])

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(serve_mock.call_count, 1)
        expected_args = ((config_mock.return_value, ("localhost", 8000)), {})
        self.assertEqual(serve_mock.call_args, expected_args)

        self.assertEqual(config_mock.call_args, ((config.SITE_YAML_PATH,), {}))
        self.assertEqual(serve_mock.return_value[0].shutdown.call_count, 0)
        self.assertEqual(serve_mock.return_value[1].join.call_count, 1)

        serve_mock.return_value[1].join.side_effect = KeyboardInterrupt
        # this time with a simulated keyboard interrupt
        result = runner.invoke(command.exhibition, ["serve"])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(serve_mock.call_count, 2)
        expected_args = ((config_mock.return_value, ("localhost", 8000)), {})
        self.assertEqual(serve_mock.call_args, expected_args)

        self.assertEqual(config_mock.call_args, ((config.SITE_YAML_PATH,), {}))
        self.assertEqual(serve_mock.return_value[0].shutdown.call_count, 1)
        self.assertEqual(serve_mock.return_value[1].join.call_count, 2)

        # Undoing the KeyboardInterrupt
        serve_mock.return_value[1].join.side_effect = None
        # Now testing w args
        config_mock.reset_mock()
        serve_mock.reset_mock()
        serve_mock.return_value[0].reset_mock()
        serve_mock.return_value[1].reset_mock()
        addr_options = ["serve", "--server", "localhost", "--port", "8001"]
        result = runner.invoke(command.exhibition, addr_options)

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(serve_mock.call_count, 1)
        expected_args = ((config_mock.return_value, ("localhost", 8001)), {})
        self.assertEqual(serve_mock.call_args, expected_args)

        self.assertEqual(serve_mock.return_value[0].shutdown.call_count, 0)
        self.assertEqual(serve_mock.return_value[1].join.call_count, 1)

    @mock.patch.object(command, "logger")
    def test_exhibition(self, log_mock):
        @command.exhibition.command()
        def test():
            pass

        # logging defaults to warn
        runner = CliRunner()
        result = runner.invoke(command.exhibition, ["test"])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(log_mock.addHandler.call_count, 1)
        self.assertEqual(log_mock.setLevel.call_count, 1)
        self.assertEqual(log_mock.setLevel.call_args, ((logging.WARN,), {}))

        # one verbose is info
        log_mock.reset_mock()
        result = runner.invoke(command.exhibition, ["-v", "test"])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(log_mock.addHandler.call_count, 1)
        self.assertEqual(log_mock.setLevel.call_count, 1)
        self.assertEqual(log_mock.setLevel.call_args, ((logging.INFO,), {}))

        # two or more verbose is debug
        log_mock.reset_mock()
        result = runner.invoke(command.exhibition, ["-vv", "test"])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(log_mock.addHandler.call_count, 1)
        self.assertEqual(log_mock.setLevel.call_count, 1)
        self.assertEqual(log_mock.setLevel.call_args, ((logging.DEBUG,), {}))

        result = runner.invoke(command.exhibition, ["-vvv", "test"])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(log_mock.addHandler.call_count, 2)
        self.assertEqual(log_mock.setLevel.call_count, 2)
        self.assertEqual(log_mock.setLevel.call_args, ((logging.DEBUG,), {}))


class CreateTestCase(TestCase):
    def _dir_cmp(self, pathA, pathB):
        dir_diff = dircmp(pathA, pathB)
        return dir_diff.left_only + dir_diff.right_only + dir_diff.funny_files + dir_diff.diff_files

    def test_create(self):
        with TemporaryDirectory() as tmpDir:
            path = pathlib.Path(tmpDir, "site")
            runner = CliRunner()
            result = runner.invoke(command.exhibition, ["create", str(path)])
            self.assertEqual(result.exit_code, 0)
            self.assertEqual(path.exists(), True)
            self.assertEqual(
                self._dir_cmp(path, pathlib.Path(exhibition.__path__[0], "data", "starter")),
                [],
            )

    def test_existing_dir(self):
        with TemporaryDirectory() as tmpDir:
            path = pathlib.Path(tmpDir, "site")
            path.mkdir()
            runner = CliRunner()
            result = runner.invoke(command.exhibition, ["create", str(path)])
            self.assertEqual(result.exit_code, 1)
            self.assertEqual(path.exists(), True)

    def test_force_with_existing_dir(self):
        with TemporaryDirectory() as tmpDir:
            path = pathlib.Path(tmpDir, "site")
            path.mkdir()
            runner = CliRunner()
            result = runner.invoke(command.exhibition, ["create", "--force", str(path)])
            self.assertEqual(result.exit_code, 0)
            self.assertEqual(path.exists(), True)
            self.assertEqual(
                self._dir_cmp(path, pathlib.Path(exhibition.__path__[0], "data", "starter")),
                [],
            )

    def test_force_without_existing_dir(self):
        with TemporaryDirectory() as tmpDir:
            path = pathlib.Path(tmpDir, "site")
            runner = CliRunner()
            result = runner.invoke(command.exhibition, ["create", "--force", str(path)])
            self.assertEqual(result.exit_code, 0)
            self.assertEqual(path.exists(), True)
            self.assertEqual(
                self._dir_cmp(path, pathlib.Path(exhibition.__path__[0], "data", "starter")),
                [],
            )



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

import io

from ruamel.yaml import YAML

SITE_YAML_PATH = "site.yaml"

yaml_parser = YAML(typ="safe")


class Config:
    """
    Configuration object that implements a dict-like interface

    If a key cannot be found in this instance, the parent :class:`Config` will
    be searched (and its parent, etc.)
    """
    def __init__(self, data=None, parent=None, node=None):
        """
        :param data:
            Can be one of a string, a file-like object, a dict-like object, or
            ``None``. The first two will be assumed as YAML
        :param parent:
            Parent :class:`Config` or ``None`` if this is the root
            configuration object
        :param node:
            The node that this object to bound to, or ``None`` if it is the
            root configuration object
        """
        assert (parent is None) == (node is None), \
            "Either both parent and node are defined or they are both None"

        self.parent = parent
        self.node = node
        self._base_config = {}

        if data:
            self.load(data)

    def load(self, data):
        """
        Load data into configutation object

        :param data:
            If a string or file-like object, ``data`` is parsed as if it were
            YAML data. If a dict-like object, ``data`` is added to the internal
            dictionary.

            Otherwise an :class:`AssertionError` exception is raised
        """
        if isinstance(data, (str, io.IOBase)):
            self._base_config.update(yaml_parser.load(data))
        elif isinstance(data, dict):
            self._base_config.update(data)
        else:
            raise AssertionError("data needs to be a string, file-like, or dict-like object")

    @classmethod
    def from_path(cls, path):
        """Load YAML data from a file"""
        with open(path) as f:
            obj = cls(f)

        return obj

    def get_name(self):
        if self.node is None:
            return SITE_YAML_PATH
        else:
            return self.node.full_path

    def __getitem__(self, key):
        try:
            return self._base_config[key]
        except KeyError as exp:
            exp_str = "Could not find %s in %s" % (key, self.get_name())
            if self.parent is None:
                raise KeyError(exp_str) from exp
            else:
                try:
                    return self.parent[key]
                except KeyError as exp_parent:
                    raise KeyError(exp_str) from exp_parent

    def __setitem__(self, key, value):
        self._base_config[key] = value

    def __contains__(self, key):
        return key in self.keys()

    def __len__(self):
        return len(list(self.keys()))

    def __iter__(self):
        return self.keys()

    def keys(self):
        _keys_set = set()
        for k in self._base_config.keys():
            _keys_set.add(k)
            yield k

        if self.parent is not None:
            for k in self.parent.keys():
                if k not in _keys_set:
                    _keys_set.add(k)
                    yield k

    def values(self):
        for k in self.keys():
            yield self[k]

    def items(self):
        for k in self.keys():
            yield (k, self[k])

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def update(self, *args, **kwargs):
        self._base_config.update(*args, **kwargs)

    def copy(self):
        klass = type(self)
        return klass(self._base_config.copy(), parent=self.parent, node=self.node)

    def __repr__(self):
        return "<%s: %s: %s>" % (self.__class__.__name__,
                                 self.get_name(),
                                 self._base_config.keys())



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

"""
External command filter

Use an external command to process a file, like so:

.. code-block:: yaml

   filter: exhibition.filters.external
   external_cmd: sed 's/this/that/g' {INPUT} > {OUTPUT}
"""

from tempfile import TemporaryDirectory
import pathlib
import subprocess

DEFAULT_GLOB = "*.*"

INPUT_NAME = "input"
OUTPUT_NAME = "output"

INPUT_KEY = "INPUT"
OUTPUT_KEY = "OUTPUT"


def content_filter(node, content):
    """
    This is the actual content filter called by :class:`exhibition.main.Node`
    on appropiate nodes.

    :param node:
        The node being rendered
    :param content:
        The content of the node, stripped of any YAML frontmatter
    """
    tmp_dir = TemporaryDirectory()
    input_file = pathlib.Path(tmp_dir.name, INPUT_NAME)
    output_file = pathlib.Path(tmp_dir.name, OUTPUT_NAME)

    cmd = node.meta["external_cmd"]
    cmd = cmd.format(**{
        INPUT_KEY: input_file,
        OUTPUT_KEY: output_file,
    })

    with input_file.open("w") as f:
        f.write(content)

    subprocess.run(cmd, shell=True)

    with output_file.open("r") as f:
        output = f.read()

    return output



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from tempfile import TemporaryDirectory
from unittest import TestCase
import hashlib
import pathlib

from exhibition.node import Node

GOOD_META = """---
thingy: 3
---
Some text
"""

BAD_END_META = """---
thingy: 3

Some text
"""

BAD_START_META = """
---
thingy: 3
---
Some text
"""

NO_META = """
Some text
"""

LONG_META = """---
%s
---
Some text
""" % "\n".join(["thing%s: 1" % i for i in range(100)])

YAML_FILE = """
thingy: 3
bob:
    - 1
    - 2
"""

YAML_WITH_IGNORE = """
ignore: "*"
"""

JSON_FILE = """
{
    "thingy": 3,
    "bob": [
        1,
        2
    ]
}
"""

BINARY_FILE = (
    b"\x00\x00\x01\x00\x02\x00\x10\x10\x10\x00\x01\x00\x04\x00(\x01\x00\x00&"
    b"\x00\x00\x00  \x10\x00\x01\x00\x04\x00\xe8\x02\x00\x00N\x01\x00\x00("
    b"\x00\x00\x00\x10\x00\x00\x00 \x00\x00\x00\x01\x00\x04\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x01\x00\x00\x08\x0c\n\x00\x16\x18\x17\x00:<;\x00"
    b"@BA\x00TV"
)

EMPTY_DIGEST = hashlib.md5().hexdigest()[:8]
BINARY_DIGEST = hashlib.md5(BINARY_FILE).hexdigest()[:8]
JSON_DIGEST = hashlib.md5(JSON_FILE.encode()).hexdigest()[:8]


class NodeTestCase(TestCase):
    def setUp(self):
        self.content_path = TemporaryDirectory()
        self.deploy_path = TemporaryDirectory()

        self.default_settings = {"content_path": self.content_path.name,
                                 "deploy_path": self.deploy_path.name}

    def tearDown(self):
        self.content_path.cleanup()
        self.deploy_path.cleanup()

    def test_repr(self):
        path = pathlib.Path(self.content_path.name, "page.html")
        path.touch()
        node = Node(path, None, meta=self.default_settings)

        self.assertEqual(repr(node), "<Node: page.html>")

    def test_full_path(self):
        path = pathlib.Path(self.content_path.name, "page.html")
        path.touch()
        node = Node(path, None, meta=self.default_settings)

        self.assertEqual(node.full_path, self.deploy_path.name)

    def test_full_path_with_parent(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "page.html")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(child_node.full_path, self.deploy_path.name + "/page.html")

    def test_full_url(self):
        path = pathlib.Path(self.content_path.name, "page.html")
        path.touch()
        node = Node(path, None, meta=self.default_settings)

        self.assertEqual(node.full_url, "/")

    def test_full_url_base_url(self):
        path = pathlib.Path(self.content_path.name, "page.html")
        path.touch()

        for base in ["/base/", "/base", "base/"]:
            self.default_settings["base_url"] = base
            node = Node(path, None, meta=self.default_settings)
            self.assertEqual(node.full_url, "/base/")

    def test_full_url_with_parent(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "page.jpg")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(child_node.full_url, "/page.jpg")

    def test_full_url_with_strip_ext(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "page.html")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(child_node.full_url, "/page")

    def test_full_url_with_index(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "index.html")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(child_node.full_url, "/")

    def test_walk(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "index.html")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(list(parent_node.walk(include_self=True)), [parent_node, child_node])
        self.assertEqual(list(parent_node.walk()), [child_node])

        self.assertEqual(list(child_node.walk(include_self=True)), [child_node])
        self.assertEqual(list(child_node.walk()), [])

    def test_render_dir(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "blog")
        child_path.mkdir()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertFalse(pathlib.Path(child_node.full_path).exists())
        child_node.render()
        self.assertTrue(pathlib.Path(child_node.full_path).exists())
        self.assertTrue(pathlib.Path(child_node.full_path).is_dir())

    def test_render_file(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "blog")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertFalse(pathlib.Path(child_node.full_path).exists())
        child_node.render()
        self.assertTrue(pathlib.Path(child_node.full_path).exists())
        self.assertTrue(pathlib.Path(child_node.full_path).is_file())

    def test_render_binary_file(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "blog")
        with child_path.open("wb") as cf:
            cf.write(BINARY_FILE)

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertFalse(pathlib.Path(child_node.full_path).exists())
        child_node.render()
        self.assertTrue(pathlib.Path(child_node.full_path).exists())
        self.assertTrue(pathlib.Path(child_node.full_path).is_file())

        with pathlib.Path(child_node.full_path).open("rb") as df:
            content = df.read()
            self.assertEqual(content, BINARY_FILE)

    def test_process_good_meta(self):
        path = pathlib.Path(self.content_path.name, "blog")
        with path.open("w") as f:
            f.write(GOOD_META)

        node = Node(path, None)
        self.assertEqual(list(node.meta.keys()), ["thingy"])
        self.assertEqual(node.content, GOOD_META[18:])

    def test_process_long_meta(self):
        path = pathlib.Path(self.content_path.name, "blog")
        with path.open("w") as f:
            f.write(LONG_META)

        node = Node(path, None)
        self.assertCountEqual(list(node.meta.keys()), ["thing%s" % i for i in range(100)])
        self.assertEqual(node.content, LONG_META[1098:])

    def test_process_bad_start_meta(self):
        path = pathlib.Path(self.content_path.name, "blog")
        with path.open("w") as f:
            f.write(BAD_START_META)

        node = Node(path, None)
        self.assertEqual(list(node.meta.keys()), [])
        self.assertEqual(node.content, BAD_START_META)

    def test_process_bad_end__meta(self):
        path = pathlib.Path(self.content_path.name, "blog")
        with path.open("w") as f:
            f.write(BAD_END_META)

        node = Node(path, None)
        self.assertEqual(list(node.meta.keys()), [])
        self.assertEqual(node.content, BAD_END_META)

    def test_process_no_meta(self):
        path = pathlib.Path(self.content_path.name, "blog")
        with path.open("w") as f:
            f.write(NO_META)

        node = Node(path, None)
        self.assertEqual(list(node.meta.keys()), [])
        self.assertEqual(node.content, NO_META)

    def test_get_content(self):
        path = pathlib.Path(self.content_path.name, "blog")
        with path.open("w") as f:
            f.write("a" * 10)

        node = Node(path, None)
        self.assertEqual(node.content, "a" * 10)

    def test_yaml_data(self):
        path = pathlib.Path(self.content_path.name, "blog.yaml")
        with path.open("w") as f:
            f.write(YAML_FILE)

        node = Node(path, None)

        self.assertEqual(node.data, {"thingy": 3, "bob": [1, 2]})

    def test_cached_data(self):
        path = pathlib.Path(self.content_path.name, "blog.yaml")
        with path.open("w") as f:
            f.write(YAML_FILE)

        node = Node(path, None)

        self.assertEqual(node.data, {"thingy": 3, "bob": [1, 2]})

        with path.open("w") as f:
            f.write("test: 1")

        self.assertEqual(node.data, {"thingy": 3, "bob": [1, 2]})

    def test_json_data(self):
        path = pathlib.Path(self.content_path.name, "blog.json")
        with path.open("w") as f:
            f.write(JSON_FILE)

        node = Node(path, None)

        self.assertEqual(node.data, {"thingy": 3, "bob": [1, 2]})

    def test_dir_data(self):
        path = pathlib.Path(self.content_path.name, "blog.json")
        path.mkdir()

        node = Node(path, None)

        self.assertEqual(node.data, None)

    def test_data_unsupported(self):
        path = pathlib.Path(self.content_path.name, "blog.xml")
        path.touch()

        node = Node(path, None)

        with self.assertRaises(KeyError):
            node.data

    def test_add_child(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "page.html")

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(parent_node.children, {"page.html": child_node})

        with self.assertRaises(AssertionError):
            parent_node.add_child(Node(pathlib.Path(self.content_path.name, "page2.html"), None))

    def test_siblings(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "page1.html")
        child2_path = pathlib.Path(self.content_path.name, "page2.html")

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child1_node = Node(child1_path, parent_node)
        child2_node = Node(child2_path, parent_node)

        self.assertEqual(child1_node.siblings, {"page2.html": child2_node})
        self.assertEqual(child2_node.siblings, {"page1.html": child1_node})

    def test_from_path(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "page1.html")
        child1_path.touch()
        child2_path = pathlib.Path(self.content_path.name, "page2.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path)
        self.assertEqual(list(parent_node.children.keys()), ["page1.html", "page2.html"])

    def test_from_path_and_ignore(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "page1.html")
        child1_path.touch()
        child2_path = pathlib.Path(self.content_path.name, "page2.html")
        child2_path.touch()
        child3_path = pathlib.Path(self.content_path.name, "picture.jpg")
        child3_path.touch()

        parent_node = Node.from_path(parent_path, meta={"ignore": "*.html"})
        self.assertEqual(list(parent_node.children.keys()), ["picture.jpg"])

    def test_from_path_and_ignore_list(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "page1.html")
        child1_path.touch()
        child2_path = pathlib.Path(self.content_path.name, "page2.html")
        child2_path.touch()
        child3_path = pathlib.Path(self.content_path.name, "picture.jpg")
        child3_path.touch()
        child4_path = pathlib.Path(self.content_path.name, "picture.gif")
        child4_path.touch()

        parent_node = Node.from_path(parent_path, meta={"ignore": ["*.html", "*.gif"]})
        self.assertEqual(list(parent_node.children.keys()), ["picture.jpg"])

    def test_from_path_with_meta(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "page1.html")
        child1_path.touch()
        child2_path = pathlib.Path(self.content_path.name, "page2.html")
        child2_path.touch()

        meta_path = pathlib.Path(self.content_path.name, "meta.yaml")
        with meta_path.open("w") as f:
            f.write("test: bob")

        parent_node = Node.from_path(parent_path)
        self.assertEqual(list(parent_node.children.keys()), ["page1.html", "page2.html"])
        self.assertEqual(parent_node.meta["test"], "bob")

    def test_from_path_meta_comes_first(self):
        parent_path = pathlib.Path(self.content_path.name)

        # not sure if default ordering for files is by creation time or by
        # name, so let's write one file here, and another after the meta.yaml
        child1_path = pathlib.Path(self.content_path.name, "1page.html")
        child1_path.touch()

        path = pathlib.Path(self.content_path.name, "meta.yaml")
        with path.open("w") as f:
            f.write(YAML_WITH_IGNORE)

        child2_path = pathlib.Path(self.content_path.name, "page2.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path, meta={})
        self.assertEqual(list(parent_node.children.keys()), [])

    def test_cache_bust_one_glob(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "bust-me.jpg")
        child1_path.touch()
        child2_path = pathlib.Path(self.content_path.name, "page.html")
        child2_path.touch()

        meta_path = pathlib.Path(self.content_path.name, "meta.yaml")
        with meta_path.open("w") as f:
            f.write("cache_bust_glob: \"*.jpg\"")

        parent_node = Node.from_path(parent_path)
        parent_node.meta.update(**self.default_settings)
        self.assertEqual(list(parent_node.children.keys()), ["bust-me.jpg", "page.html"])

        child1_node = parent_node.children["bust-me.jpg"]
        child2_node = parent_node.children["page.html"]

        self.assertEqual(child1_node.full_path,
                         str(pathlib.Path(self.deploy_path.name,
                                          "bust-me.{}.jpg".format(EMPTY_DIGEST))))
        self.assertEqual(child2_node.full_path,
                         str(pathlib.Path(self.deploy_path.name, "page.html")))

        self.assertEqual(child1_node.full_url, "/bust-me.{}.jpg".format(EMPTY_DIGEST))
        self.assertEqual(child2_node.full_url, "/page")

    def test_cache_bust_multi_glob(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "bust-me.jpg")
        child1_path.touch()
        child2_path = pathlib.Path(self.content_path.name, "page.html")
        child2_path.touch()

        meta_path = pathlib.Path(self.content_path.name, "meta.yaml")
        with meta_path.open("w") as f:
            f.write("cache_bust_glob:\n  - \"*.jpeg\"\n  - \"*.jpg\"")

        parent_node = Node.from_path(parent_path)
        parent_node.meta.update(**self.default_settings)
        self.assertEqual(list(parent_node.children.keys()), ["bust-me.jpg", "page.html"])

        child1_node = parent_node.children["bust-me.jpg"]
        child2_node = parent_node.children["page.html"]

        self.assertEqual(child1_node.full_path,
                         str(pathlib.Path(self.deploy_path.name,
                                          "bust-me.{}.jpg".format(EMPTY_DIGEST))))
        self.assertEqual(child2_node.full_path,
                         str(pathlib.Path(self.deploy_path.name, "page.html")))

        self.assertEqual(child1_node.full_url, "/bust-me.{}.jpg".format(EMPTY_DIGEST))
        self.assertEqual(child2_node.full_url, "/page")

    def test_cache_bust_binary_file(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "bust-me.jpg")
        with child_path.open("wb") as f:
            f.write(BINARY_FILE)

        parent_node = Node.from_path(parent_path)
        parent_node.meta.update(**self.default_settings)
        parent_node.meta["cache_bust_glob"] = "*"
        self.assertEqual(list(parent_node.children.keys()), ["bust-me.jpg"])

        child_node = parent_node.children["bust-me.jpg"]

        self.assertEqual(child_node.full_path,
                         str(pathlib.Path(self.deploy_path.name,
                                          "bust-me.{}.jpg".format(BINARY_DIGEST))))

        self.assertEqual(child_node.full_url, "/bust-me.{}.jpg".format(BINARY_DIGEST))

    def test_cache_bust_text_file(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "bust-me.jpg")
        with child_path.open("w") as f:
            f.write(JSON_FILE)

        parent_node = Node.from_path(parent_path)
        parent_node.meta.update(**self.default_settings)
        parent_node.meta["cache_bust_glob"] = "*"
        self.assertEqual(list(parent_node.children.keys()), ["bust-me.jpg"])

        child_node = parent_node.children["bust-me.jpg"]

        self.assertEqual(child_node.full_path,
                         str(pathlib.Path(self.deploy_path.name,
                                          "bust-me.{}.jpg".format(JSON_DIGEST))))

        self.assertEqual(child_node.full_url, "/bust-me.{}.jpg".format(JSON_DIGEST))

    def test_get_from_path_relative_str_on_leaf(self):
        parent_path = pathlib.Path(self.content_path.name)
        pathlib.Path(self.content_path.name, "images").mkdir()
        child1_path = pathlib.Path(self.content_path.name, "images", "bust-me.jpg")
        child1_path.touch()
        pathlib.Path(self.content_path.name, "pages").mkdir()
        child2_path = pathlib.Path(self.content_path.name, "pages", "page.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path)

        child_node = parent_node.children["images"].children["bust-me.jpg"]
        target_node = parent_node.children["pages"].children["page.html"]

        self.assertEqual(child_node.get_from_path("../pages/page.html"), target_node)

    def test_get_from_path_relative_pathlib_on_leaf(self):
        parent_path = pathlib.Path(self.content_path.name)
        pathlib.Path(self.content_path.name, "images").mkdir()
        child1_path = pathlib.Path(self.content_path.name, "images", "bust-me.jpg")
        child1_path.touch()
        pathlib.Path(self.content_path.name, "pages").mkdir()
        child2_path = pathlib.Path(self.content_path.name, "pages", "page.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path)

        child_node = parent_node.children["images"].children["bust-me.jpg"]
        target_node = parent_node.children["pages"].children["page.html"]

        self.assertEqual(child_node.get_from_path(pathlib.Path("..", "pages", "page.html")),
                         target_node)

    def test_get_from_path_relative_str_on_dir(self):
        parent_path = pathlib.Path(self.content_path.name)
        pathlib.Path(self.content_path.name, "images").mkdir()
        child1_path = pathlib.Path(self.content_path.name, "images", "bust-me.jpg")
        child1_path.touch()
        pathlib.Path(self.content_path.name, "pages").mkdir()
        child2_path = pathlib.Path(self.content_path.name, "pages", "page.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path)

        child_node = parent_node.children["images"]
        target_node = parent_node.children["pages"].children["page.html"]

        self.assertEqual(child_node.get_from_path("../pages/page.html"), target_node)

    def test_get_from_path_relative_pathlib_on_dir(self):
        parent_path = pathlib.Path(self.content_path.name)
        pathlib.Path(self.content_path.name, "images").mkdir()
        child1_path = pathlib.Path(self.content_path.name, "images", "bust-me.jpg")
        child1_path.touch()
        pathlib.Path(self.content_path.name, "pages").mkdir()
        child2_path = pathlib.Path(self.content_path.name, "pages", "page.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path)

        child_node = parent_node.children["images"]
        target_node = parent_node.children["pages"].children["page.html"]

        self.assertEqual(child_node.get_from_path(pathlib.Path("..", "pages", "page.html")),
                         target_node)

    def test_get_from_path_relative_str_with_dot(self):
        parent_path = pathlib.Path(self.content_path.name)
        pathlib.Path(self.content_path.name, "images").mkdir()
        child1_path = pathlib.Path(self.content_path.name, "images", "bust-me.jpg")
        child1_path.touch()
        pathlib.Path(self.content_path.name, "pages").mkdir()
        child2_path = pathlib.Path(self.content_path.name, "pages", "page.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path)

        child_node = parent_node.children["pages"]
        target_node = parent_node.children["pages"].children["page.html"]

        self.assertEqual(child_node.get_from_path("./page.html"), target_node)

    def test_get_from_path_relative_pathlib_with_dot(self):
        parent_path = pathlib.Path(self.content_path.name)
        pathlib.Path(self.content_path.name, "images").mkdir()
        child1_path = pathlib.Path(self.content_path.name, "images", "bust-me.jpg")
        child1_path.touch()
        pathlib.Path(self.content_path.name, "pages").mkdir()
        child2_path = pathlib.Path(self.content_path.name, "pages", "page.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path)

        child_node = parent_node.children["pages"]
        target_node = parent_node.children["pages"].children["page.html"]

        self.assertEqual(child_node.get_from_path(pathlib.Path(".", "page.html")),
                         target_node)

    def test_get_from_path_absolute_str(self):
        parent_path = pathlib.Path(self.content_path.name)
        pathlib.Path(self.content_path.name, "images").mkdir()
        child1_path = pathlib.Path(self.content_path.name, "images", "bust-me.jpg")
        child1_path.touch()
        pathlib.Path(self.content_path.name, "pages").mkdir()
        child2_path = pathlib.Path(self.content_path.name, "pages", "page.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path)

        child_node = parent_node.children["images"].children["bust-me.jpg"]
        target_node = parent_node.children["pages"].children["page.html"]

        self.assertEqual(child_node.get_from_path("/pages/page.html"), target_node)

    def test_get_from_path_absolute_pathlib(self):
        parent_path = pathlib.Path(self.content_path.name)
        pathlib.Path(self.content_path.name, "images").mkdir()
        child1_path = pathlib.Path(self.content_path.name, "images", "bust-me.jpg")
        child1_path.touch()
        pathlib.Path(self.content_path.name, "pages").mkdir()
        child2_path = pathlib.Path(self.content_path.name, "pages", "page.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path)

        child_node = parent_node.children["images"].children["bust-me.jpg"]
        target_node = parent_node.children["pages"].children["page.html"]

        self.assertEqual(child_node.get_from_path(pathlib.Path("/pages/page.html")), target_node)

    def test_get_from_path_not_existing_from_str(self):
        parent_path = pathlib.Path(self.content_path.name)
        pathlib.Path(self.content_path.name, "images").mkdir()
        child1_path = pathlib.Path(self.content_path.name, "images", "bust-me.jpg")
        child1_path.touch()
        pathlib.Path(self.content_path.name, "pages").mkdir()
        child2_path = pathlib.Path(self.content_path.name, "pages", "page.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path)

        child_node = parent_node.children["images"].children["bust-me.jpg"]

        with self.assertRaises(OSError):
            child_node.get_from_path("../not-a-page.html")

    def test_get_from_path_not_existing_from_pathlib(self):
        parent_path = pathlib.Path(self.content_path.name)
        pathlib.Path(self.content_path.name, "images").mkdir()
        child1_path = pathlib.Path(self.content_path.name, "images", "bust-me.jpg")
        child1_path.touch()
        pathlib.Path(self.content_path.name, "pages").mkdir()
        child2_path = pathlib.Path(self.content_path.name, "pages", "page.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path)

        child_node = parent_node.children["images"].children["bust-me.jpg"]

        with self.assertRaises(OSError):
            child_node.get_from_path(pathlib.Path("..", "not-a-page.html"))

    def test_root_node_is_kept(self):
        parent_path = pathlib.Path(self.content_path.name)
        pathlib.Path(self.content_path.name, "images").mkdir()
        pathlib.Path(self.content_path.name, "pages").mkdir()
        child_path = pathlib.Path(self.content_path.name, "pages", "page.html")
        child_path.touch()

        parent_node = Node.from_path(parent_path)

        self.assertEqual(parent_node.root_node, parent_node)
        self.assertEqual(parent_node.children["pages"].root_node, parent_node)
        self.assertEqual(parent_node.children["pages"].children["page.html"].root_node, parent_node)
        self.assertEqual(parent_node.children["pages"].children["page.html"].parent.root_node,
                         parent_node)
        self.assertEqual(
            parent_node.children["pages"].children["page.html"].parent.parent.root_node,
            parent_node
        )
        self.assertEqual(
            parent_node.children["pages"].children["page.html"].parent.parent.root_node,
            parent_node
        )

    def test_strip_exts_default(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "blog.html")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(child_node.strip_exts, [".html"])
        self.assertEqual(child_node.full_url, "/blog")

    def test_strip_exts_custom(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "blog.tw2")
        child_path.touch()

        settings = {"strip_exts": [".tw2"]}
        settings.update(self.default_settings)
        parent_node = Node(parent_path, None, meta=settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(child_node.strip_exts, [".tw2"])
        self.assertEqual(child_node.full_url, "/blog")

    def test_strip_exts_multiple(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "blog.bluh")
        child1_path.touch()
        child2_path = pathlib.Path(self.content_path.name, "story.tw2")
        child2_path.touch()

        settings = {"strip_exts": [".tw2", ".bluh"]}
        settings.update(self.default_settings)
        parent_node = Node(parent_path, None, meta=settings)
        child1_node = Node(child1_path, parent_node)
        child2_node = Node(child2_path, parent_node)

        self.assertEqual(child1_node.strip_exts, [".tw2", ".bluh"])
        self.assertEqual(child1_node.full_url, "/blog")

        self.assertEqual(child2_node.strip_exts, [".tw2", ".bluh"])
        self.assertEqual(child2_node.full_url, "/story")

    def test_strip_exts_not_a_list(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "blog.html")
        child_path.touch()

        settings = {"strip_exts": ".html"}
        settings.update(self.default_settings)
        parent_node = Node(parent_path, None, meta=settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(child_node.strip_exts, [".html"])
        self.assertEqual(child_node.full_url, "/blog")

    def test_strip_exts_disabled(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "blog.html")
        child_path.touch()

        settings = {"strip_exts": []}
        settings.update(self.default_settings)
        parent_node = Node(parent_path, None, meta=settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(child_node.strip_exts, [])
        self.assertEqual(child_node.full_url, "/blog.html")

    def test_index_file_default(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "index.html")
        child1_path.touch()
        child2_path = pathlib.Path(self.content_path.name, "blog.html")
        child2_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child1_node = Node(child1_path, parent_node)
        child2_node = Node(child2_path, parent_node)

        self.assertEqual(child1_node.index_file, "index.html")
        self.assertEqual(child1_node.full_url, "/")
        self.assertEqual(child2_node.index_file, "index.html")
        self.assertEqual(child2_node.full_url, "/blog")

    def test_index_file_custom(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "index.html")
        child1_path.touch()
        child2_path = pathlib.Path(self.content_path.name, "blog.html")
        child2_path.touch()

        settings = {"index_file": "blog.html"}
        settings.update(self.default_settings)
        parent_node = Node(parent_path, None, meta=settings)
        child1_node = Node(child1_path, parent_node)
        child2_node = Node(child2_path, parent_node)

        self.assertEqual(child1_node.index_file, "blog.html")
        self.assertEqual(child1_node.full_url, "/index")
        self.assertEqual(child2_node.index_file, "blog.html")
        self.assertEqual(child2_node.full_url, "/")



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

"""
Documentation for this module can be found in :doc:`commandline`
"""

import logging
import pathlib
import shutil

import click

from exhibition import __version__, config, utils
import exhibition as exhib_module

logger = logging.getLogger("exhibition")


@click.group()
@click.version_option(version=__version__)
@click.option("-v", "--verbose", count=True,
              help="Verbose output, can be used multiple times to increase logging level")
def exhibition(verbose):
    """
    A Python static site generator
    """
    logger.addHandler(logging.StreamHandler())
    if verbose > 1:
        logger.setLevel(logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARN)


@exhibition.command(short_help="Generate site")
def gen():
    """
    Generate site from content_path
    """
    settings = config.Config.from_path(config.SITE_YAML_PATH)
    utils.gen(settings)


@exhibition.command(short_help="Serve site locally")
@click.option("-s", "--server", default="localhost", help="Hostname to serve the site at.")
@click.option("-p", "--port", default=8000, type=int, help="Port to serve the site at.")
def serve(server, port):
    """
    Serve files from deploy_path as a webserver would
    """
    settings = config.Config.from_path(config.SITE_YAML_PATH)
    server_address = (server, port)
    httpd, thread = utils.serve(settings, server_address)

    try:
        thread.join()
    except (KeyboardInterrupt, SystemExit):
        httpd.shutdown()


@exhibition.command(short_help="Create a starter site")
@click.argument("destination", metavar="PATH")
@click.option("-f", "--force", is_flag=True, help="overwrite destination if it exists")
def create(destination, force):
    """
    Generate a starter site with a basic configuration for website.
    """
    starter = pathlib.Path(exhib_module.__path__[0], "data", "starter")

    if force:
        shutil.rmtree(destination, ignore_errors=True)

    if pathlib.Path(destination).exists():
        raise click.ClickException("Directory '{}' exists, aborting".format(destination))

    shutil.copytree(starter, destination)



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from collections import OrderedDict
from functools import cached_property
from importlib import import_module
import hashlib
import pathlib

from ruamel.yaml import YAML

from .config import Config

yaml_parser = YAML(typ="safe")

DATA_EXTRACTORS = {
    ".yaml": yaml_parser.load,
    ".json": yaml_parser.load,
}

DEFAULT_STRIP_EXTS = [".html"]
DEFAULT_INDEX_FILE = "index.html"


class Node:
    """
    A node represents a file or directory
    """
    _meta_names = ["meta.yaml", "meta.yml"]

    _meta_header = "---\n"
    _meta_footer = "---\n"

    _dir_mode = 0o755
    _file_mode = 0o644

    def __init__(self, path, parent, meta=None):
        """
        :param path:
            A :class:`pathlib.Path` that is either the ``content_path`` or a
            child of it.
        :param parent:
            Either another :class:`Node` or ``None``
        :param meta:
            A dict-like object that will be passed to a :class:`Config`
            instance
        """
        self.path_obj = path
        self.parent = parent
        self.children = OrderedDict()

        self.is_leaf = self.path_obj.is_file()

        try:
            parent_meta = self.parent.meta
        except AttributeError:
            parent_meta = None

        self.__meta = Config({}, parent=parent_meta, node=self.parent)
        if meta:
            self.__meta.update(meta)

        if self.parent:
            self.parent.add_child(self)
            self.root_node = self.parent.root_node
        else:
            self.root_node = self

    @classmethod
    def from_path(cls, path, parent=None, meta=None):
        """
        Given a :class:`pathlib.Path`, create a Node from that path as well as
        any children. Children are loaded in Unicode codepoint order - this
        order is preserved in ``Node.children`` if you're unsure what that
        means.

        If the path is not a file or a dir, an :class:`AssertionError` is
        raised

        :param path:
            A :class:`pathlib.Path` that is either the ``content_path`` or a
            child of it.
        :param parent:
            Either another :class:`Node` or ``None``
        :param meta:
            A dict-like object that will be passed to a :class:`Config`
            instance
        """
        # path should be a pathlib object
        assert path.is_file() or path.is_dir()

        node = cls(path, parent=parent, meta=meta)

        if path.is_dir():
            children = []

            dir_files = sorted(path.iterdir(), key=lambda p: p.name)
            for child in dir_files:
                if child.name in cls._meta_names and child.is_file():
                    with child.open() as co:
                        node.meta.load(co)
                else:
                    children.append(child)

            for child in children:
                ignored = False
                globs = node.meta.get("ignore", [])

                if not isinstance(globs, (list, tuple)):
                    globs = [globs]

                for glob in globs:
                    if child in path.glob(glob):
                        ignored = True
                        break
                if not ignored:
                    cls.from_path(child, node)

        return node

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.path_obj.name)

    def get_from_path(self, path):
        """
        Given a relative or absolute path, return the :class:`Node` that
        represents that path.

        :param path:
            A :class:`str` or :class:`pathlib.Path`
        """
        if not isinstance(path, pathlib.PurePath):
            path = pathlib.PurePath(path)

        if path.is_absolute():
            found_node = self.root_node
            parts = path.parts[1:]
        else:
            if self.is_leaf:
                found_node = self.parent
            else:
                found_node = self
            parts = path.parts

        try:
            for part in parts:
                if part == "..":
                    found_node = found_node.parent
                else:
                    found_node = found_node.children[part]
        except KeyError as exp:
            raise OSError("{} could not find {}".format(self, exp.args)) from exp

        return found_node

    @property
    def full_path(self):
        """
        Full path of node when deployed
        """
        if self.parent is None:
            return self.meta["deploy_path"]
        else:
            if self.cache_bust:
                suffixes = "".join(self.path_obj.suffixes)
                pth = self.path_obj.with_suffix(".{}{}".format(self.cache_bust, suffixes))
                name = pth.name
            else:
                name = self.path_obj.name
            return str(pathlib.Path(self.parent.full_path, name))

    @property
    def full_url(self):
        """Get full URL for node, including trailing slash"""
        if self.parent is None:
            base_url = self.meta.get("base_url", "/")
            if not base_url.startswith("/"):
                base_url = "/" + base_url
            if not base_url.endswith("/"):
                base_url = base_url + "/"

            return base_url
        elif self.is_leaf:
            if self.path_obj.name == self.index_file:
                name = ""
            elif self.path_obj.suffix in self.strip_exts:
                name = self.path_obj.stem
            elif self.cache_bust:
                suffixes = "".join(self.path_obj.suffixes)
                pth = self.path_obj.with_suffix(".{}{}".format(self.cache_bust, suffixes))
                name = pth.name
            else:
                name = self.path_obj.name

            return "".join([self.parent.full_url, name])
        else:
            return "".join([self.parent.full_url, self.path_obj.name, "/"])

    def walk(self, include_self=False):
        """Walk through Node tree"""
        if include_self:
            yield self

        for child in self.children.values():
            yield child
            for grandchild in child.walk():
                yield grandchild

    def render(self):
        """
        Process node and either create the directory or write contents of file
        to ``deploy_path``
        """
        if not self.is_leaf:
            pathlib.Path(self.full_path).mkdir(self._dir_mode)
            return

        file_obj = pathlib.Path(self.full_path)
        file_obj.touch(self._file_mode)

        with file_obj.open("w" if type(self.content) is str else "wb") as fo:
            fo.write(self.content)

    @cached_property
    def content(self):
        """
        Get the actual content of the Node

        If ``filter`` has been specified in :attr:`meta`, that filter will be
        used to further process the content.
        """
        content_filter = self.meta.get("filter")
        with self.path_obj.open("rb") as file_obj:
            file_obj.seek(self.__content_start)
            content = file_obj.read()
            try:
                content = content.decode("utf-8")
            except UnicodeDecodeError:
                return content

            if content_filter is not None:
                filter_module = import_module(content_filter)
                globs = self.meta.get("filter_glob", filter_module.DEFAULT_GLOB)
                if not isinstance(globs, (list, tuple)):
                    globs = [globs]
                for filter_glob in globs:
                    if self.path_obj in self.path_obj.parent.glob(filter_glob):
                        return filter_module.content_filter(self, content)
        return content

    @cached_property
    def meta(self):
        """
        Configuration object

        Finds and processes the YAML front matter at the top of a file

        If the file does not start with ``---\\n``, then it's assumed the file
        does not contain any meta YAML for us to process
        """
        self.__content_start = 0
        if not self.is_leaf:
            return self.__meta

        found_header = False
        with self.path_obj.open("rb") as file_obj:
            while True:
                data = file_obj.read(100)
                try:
                    data = data.decode("utf-8")
                except UnicodeDecodeError:
                    return self.__meta

                if data == '':
                    # we've run out of file, either we didn't find a header or
                    # we're missing a footer
                    return self.__meta
                elif not found_header:
                    if data.startswith(self._meta_header):
                        found_header = True
                        found_meta = data[len(self._meta_header):]
                    else:
                        # if our token is not the first thing in the file, then
                        # it's not for us
                        return self.__meta
                else:
                    found_meta += data

                if self._meta_footer in found_meta:
                    idx = found_meta.index(self._meta_footer)
                    found_meta = found_meta[:idx]
                    break

        self.__meta.load(found_meta)
        self.__content_start = len(self._meta_header) + len(found_meta) + len(self._meta_footer)

        return self.__meta

    @property
    def marks(self):
        """
        Marked sections from content
        """
        if not hasattr(self, "_marks"):
            self._marks = {}
            # make sure that _marks gets populated
            self.content

        return self._marks

    @cached_property
    def data(self):
        """Extracts data from contents of file

        For example, a YAML file
        """
        if self.path_obj.is_dir():
            return

        func = DATA_EXTRACTORS[self.path_obj.suffix]
        data = func(self.content)

        return data

    def add_child(self, child):
        """
        Add a child to the current Node

        If the child doesn't already have its :attr:`parent` set to this Node,
        then an :class:`AssertionError` is raised.
        """
        assert child.parent == self
        self.children[child.path_obj.name] = child

    @property
    def siblings(self):
        """Returns all children of the parent Node, except for itself"""
        return {k: v for k, v in self.parent.children.items() if v is not self}

    @cached_property
    def cache_bust(self):
        cache_bust_version = None
        globs = self.meta.get("cache_bust_glob", [])
        if not isinstance(globs, (list, tuple)):
            globs = [globs]
        for cache_bust_glob in globs:
            if self.path_obj in self.path_obj.parent.glob(cache_bust_glob):
                hasher = hashlib.md5()
                content = self.content
                if isinstance(content, str):
                    # content needs to be bytes just for this bit
                    content = content.encode("utf-8")
                hasher.update(content)

                cache_bust_version = hasher.hexdigest()[:8]
                break

        return cache_bust_version

    @property
    def strip_exts(self):
        strip_exts = self.meta.get("strip_exts", DEFAULT_STRIP_EXTS)
        if not isinstance(strip_exts, (list, tuple)):
            strip_exts = [strip_exts]

        return strip_exts

    @property
    def index_file(self):
        index_file = self.meta.get("index_file", DEFAULT_INDEX_FILE)

        return index_file



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from http.client import HTTPConnection
from tempfile import TemporaryDirectory
from unittest import TestCase
import pathlib

from exhibition.config import Config
from exhibition.utils import serve

INDEX_CONTENTS = """<html>
    <head>
        <title>Hello</title>
    </head>
    <body>
        Hello world!
    </body>
</html>
"""

BLOG_INDEX_CONTENTS = """<html>
    <head>
        <title>Blog</title>
    </head>
    <body>
        Goodbye world!
    </body>
</html>
"""

PAGE_CONTENTS = """<html>
    <head>
        <title>Page 1</title>
    </head>
    <body>
        This is a page
    </body>
</html>
"""

CSS_CONTENTS = """// CSS file
html, body {
    color: black;
    background-color: white;
}
"""


class ServeTestCase(TestCase):
    def setUp(self):
        self.tmp_dir = TemporaryDirectory()

        with pathlib.Path(self.tmp_dir.name, "index.html").open("w") as index:
            index.write(INDEX_CONTENTS)

        with pathlib.Path(self.tmp_dir.name, "page.html").open("w") as page:
            page.write(PAGE_CONTENTS)

        with pathlib.Path(self.tmp_dir.name, "style.css").open("w") as css:
            css.write(CSS_CONTENTS)

        self.blog_dir = pathlib.Path(self.tmp_dir.name, "blog")
        self.blog_dir.mkdir()

        with pathlib.Path(self.blog_dir, "index.html").open("w") as index:
            index.write(BLOG_INDEX_CONTENTS)

        self.client = HTTPConnection("localhost", "8000")

    def tearDown(self):
        self.tmp_dir.cleanup()
        self.client.close()
        self.server.shutdown()
        self.server.server_close()

    def get_server(self, settings):
        server_address = ("localhost", 8000)
        httpd, thread = serve(settings, server_address)
        self.server = httpd

    def test_fetch_file(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/style.css")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, CSS_CONTENTS.encode())
        headers = dict(response.getheaders())
        self.assertEqual(headers["Cache-Control"], "no-store")

    def test_fetch_file_with_get_params(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/style.css?something=1")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, CSS_CONTENTS.encode())
        headers = dict(response.getheaders())
        self.assertEqual(headers["Cache-Control"], "no-store")

    def test_fetch_file_with_fragment(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/style.css#something")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, CSS_CONTENTS.encode())
        headers = dict(response.getheaders())
        self.assertEqual(headers["Cache-Control"], "no-store")

    def test_fetch_file_with_prefix(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name,
                           "base_url": "/bob/"})
        self.get_server(settings)

        self.client.request("GET", "/bob/style.css")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, CSS_CONTENTS.encode())

        self.client.request("GET", "/style.css")
        response = self.client.getresponse()
        self.assertEqual(response.status, 404)

    def test_index(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, INDEX_CONTENTS.encode())
        headers = dict(response.getheaders())
        self.assertEqual(headers["Cache-Control"], "no-store")

    def test_404(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/not-existing.html")
        response = self.client.getresponse()
        self.assertEqual(response.status, 404)
        headers = dict(response.getheaders())
        self.assertEqual(headers["Cache-Control"], "no-store")

    def test_stripped_extension(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/page")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, PAGE_CONTENTS.encode())

    def test_custom_stripped_extension(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name,
                           "strip_exts": ".css"})
        self.get_server(settings)

        self.client.request("GET", "/style")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, CSS_CONTENTS.encode())

        self.client.request("GET", "/page")
        response = self.client.getresponse()
        self.assertEqual(response.status, 404)

    def test_stripped_extension_and_trailing_slash(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/page/")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, PAGE_CONTENTS.encode())

    def test_stripped_extension_with_extension(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/page.html")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, PAGE_CONTENTS.encode())

    def test_stripped_extension_not_exists(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/unknown")
        response = self.client.getresponse()
        self.assertEqual(response.status, 404)

    def test_dir_index(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/blog/")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, BLOG_INDEX_CONTENTS.encode())

    def test_dir_without_index(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "content_path": self.tmp_dir.name})
        index = pathlib.Path(self.blog_dir, "index.html")
        index.unlink()
        self.assertFalse(index.exists())
        self.get_server(settings)

        self.client.request("GET", "/blog/")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from http.server import HTTPServer, SimpleHTTPRequestHandler
import logging
import pathlib
import shutil
import threading

from .node import Node

logger = logging.getLogger("exhibition")


def gen(settings):
    """
    Generate site

    Deletes ``deploy_path`` first.
    """
    shutil.rmtree(settings["deploy_path"], True)
    root_node = Node.from_path(pathlib.Path(settings["content_path"]), meta=settings)

    for item in root_node.walk(True):
        logger.info("Rendering %s", item.full_url)
        item.render()


class ExhibitionBaseHTTPRequestHandler(SimpleHTTPRequestHandler):
    def _sanitise_path(self, path):
        """ Strip leading and trailing / as well as base_url, if preset """
        path = path.split('?', 1)[0]
        path = path.split('#', 1)[0]
        path = path.strip("/")
        if self._settings.get("base_url"):
            base = self._settings["base_url"].strip("/")
            if not path.startswith(base):
                return
            path = path.lstrip(base).strip("/")

        return path

    def translate_path(self, path):
        path = self._sanitise_path(path)
        root_node = Node.from_path(pathlib.Path(self._settings["content_path"]),
                                   meta=self._settings)

        try:
            node = root_node.get_from_path(pathlib.PurePath(path).parent or path)
        except (OSError, TypeError):
            return ""

        path = pathlib.Path(self._settings["deploy_path"], path)

        if not (path.exists() or path.suffix):
            for ext in node.strip_exts:
                new_path = path.with_suffix(ext)
                if new_path.exists():
                    return str(new_path)
        elif path.is_dir():
            new_path = pathlib.Path(path, node.index_file)
            if new_path.exists():
                return str(new_path)

        return str(path)

    def end_headers(self):
        self.send_header("Cache-Control", "no-store")
        return super().end_headers()


def serve(settings, server_address):
    """
    Serves the generated site from ``deploy_path``

    Respects settings like ``base_url`` if present.
    """
    logger = logging.getLogger("exhibition.server")

    # this is quite ewwww, but whatever.
    class ExhibitionHTTPRequestHandler(ExhibitionBaseHTTPRequestHandler):
        _settings = settings

    httpd = HTTPServer(server_address, ExhibitionHTTPRequestHandler)

    logger.warning("Listening on http://%s:%s", *server_address)
    t = threading.Thread(target=httpd.serve_forever, daemon=True)
    t.start()

    return (httpd, t)
